//----------------------------------------------------------------------------
//	SOURCE:	MapProj.c
//	AUTHOR:	Ian Taylor
//	NOTES:	Projection implementations for Map demo application
//----------------------------------------------------------------------------

#include "Map.h"
#include "MapProj.h"

// Intrinsic mathematical functions
#pragma intrinsic(fabs, sqrt, exp, pow, log, sin, cos, tan, atan, atan2)

// Constants
#define pi					(float)M_PI
#define sin_30				0.5F
#define sin_45				(float)M_SIN_45
#define sin_60				(float)M_SIN_60
#define cos_30				(float)M_SIN_60
#define cos_45				(float)M_SIN_45
#define cos_60				0.5F

// Trig and stuff
#define dms2deg(d, m, s)	((d) + (m) / 60.0F + (s) / 3600.0F)
#define deg2rad(x)			(((x) * pi) / 180)
#define rabs(x)				((float)fabs(x))
#define rln(x)				((float)log(x))
#define rsin(x)				((float)sin(x))
#define rcos(x)				((float)cos(x))
#define rtan(x)				((float)tan(x))
#define rexp(x)				((float)exp(x))
#define rpow(x,y)			((float)pow((x), (y)))

local float rclamp(float x, float floor, float ceil)
{
	// Clamp a float value to a range
	if (x < floor)
		return floor;
	if (x > ceil)
		return ceil;
	return x;
}

local int rsgn(float x)
{
	// Calculate the signum of a number
	return (x < 0) ? -1 : (x > 0) ? 1 : 0;
}

local float rsqr(float x)
{
	// Calculate the square of a number
	return x * x;
}

local float rsqrt(float x)
{
	// Calculate the square root of a number
	ASSERT(x > 0);
	return (float)sqrt(x);
}

local float rsqrts(float x)
{
	// Calculate the square root of a number (zero if non-positive)
	return (x > 0) ? (float)sqrt(x) : 0;
}

local float rmin(float x, float y)
{
	// Return the minimum of two reals
	if (x < y)
		return x;
	return y;
}

local float rmax(float x, float y)
{
	// Return the maximum of two reals
	if (x > y)
		return x;
	return y;
}

local float rasin(float x)
{
	// Arcsin via arctan
	float y;
	ASSERT((x >= -1) && (x <= 1));
	y = 1 - x * x;
	ASSERT(y >= 0);
	y = (float)atan2(x, sqrt(y));
	ASSERT((y >= (-pi / 2)) && (y <= (pi / 2)));
	return y;
}

local float racos(float x)
{
	// Arccos via arctan
	float y;
	ASSERT((x >= -1) && (x <= 1));
	y = 1 - x * x;
	ASSERT(y >= 0);
	y = (float)atan2(sqrt(y), x);
	ASSERT((y >= 0) && (y <= pi));
	return y;
}

local float ratan(float y, float x)
{
	// Arctan
	float z = (float)atan2(y, x);
	ASSERT((z >= -pi) && (z <= pi));
	return z;
}

local float SolveThetaPlusSinTheta(float rhs, float theta)
{
	// Solve: 'theta + sin(theta) = rhs' w.r.t 'theta'
	int i;
	for (i = 0; i < 8; i++)
	{
		const float a = theta + rsin(theta) - rhs;
		const float b = 1 + rcos(theta);
		ASSERT(b > 0);
		theta -= a / b;
		if (rabs(a) < 0.000001F)
			break;
	}
	ASSERT(rabs(theta + rsin(theta) - rhs) < 0.000002F);
	return theta;
}

local float SolveThetaPlusSinThetaCosThetaPlusTwoSinTheta(float rhs, float theta)
{
	// Solve: 'theta + sin(theta).cos(theta) + 2.sin(theta) = rhs' w.r.t 'theta'
	int i;
	for (i = 0; i < 8; i++)
	{
		const float s = rsin(theta);
		const float c = rcos(theta);
		const float a = theta + s * (c + 2) - rhs;
		const float b = 1 + c * (c + 2) - s * s;
		ASSERT(b > 0);
		theta -= a / b;
		if (rabs(a) < 0.000001F)
			break;
	}
	ASSERT(rabs(theta + rsin(theta) * (rcos(theta) + 2) - rhs) < 0.000002F);
	return theta;
}

local float SolveTwoThetaPlusSinTwoThetaMinusTwoSinTheta(float rhs, float theta)
{
	// Solve: '2.theta + sin(2.theta) - 2.sin(theta) = rhs' w.r.t 'theta'
	//		= '2.theta + 2.sin(theta).cos(theta) - 2.sin(theta) = rhs'
	//		= 'theta + sin(theta).(cos(theta) - 1) = rhs/2'
	int i;
	for (i = 0; i < 8; i++)
	{
		const float s = rsin(theta);
		const float c = rcos(theta);
		const float a = theta + s * (c - 1) - rhs / 2;
		const float b = 1 - s * s + c * (c - 1);
		ASSERT(b > 0);
		theta -= a / b;
		if (rabs(a) < 0.000001F)
			break;
	}
	ASSERT(rabs(2 * theta + rsin(2 * theta) - 2 * rsin(theta) - rhs) < 0.000002F);
	return theta;
}

local float SolveMollweide(float phi)
{
	// Solve: '2.theta + sin(2.theta) = pi.sin(phi)' w.r.t 'theta'
	if (rabs(phi) > 1.5707963F)
		return rsgn(phi) * pi / 2;
	return SolveThetaPlusSinTheta(pi * rsin(phi), 2 * rasin(2 * phi / pi)) / 2;
}

local float SolveNell(float phi)
{
	// Solve: 'theta + sin(theta) = 2.sin(phi)' w.r.t 'theta'
	const float phi2 = rsqr(phi);
	const float theta = phi * (1.00371F + phi2 * (-0.0935382F + phi2 * -0.011412F));
	return SolveThetaPlusSinTheta(2 * rsin(phi), theta);
}

local float SolveEckert4(float phi)
{
	// Solve: 'theta + sin(theta) = (2 + pi/2).sin(phi)' w.r.t 'theta'
	const float phi2 = rsqr(phi);
	const float theta = phi * 0.895168F + phi2 * (0.0218849F + phi2 * 0.00826809F);
	return SolveThetaPlusSinThetaCosThetaPlusTwoSinTheta((2 + pi / 2) * rsin(phi), theta);
}

local float SolveEckert6(float phi)
{
	// Solve: 'theta + sin(theta) = (1 + pi/2).sin(phi)' w.r.t 'theta'
	return SolveThetaPlusSinTheta((1 + pi / 2) * rsin(phi), phi);
}

local float SolveWagner(float rhs)
{
	// Solve: '2.theta + sin(2.theta) = rhs' w.r.t 'theta'
	return SolveThetaPlusSinTheta(rhs, rhs / 2) / 2;
}

local float SolvePutnins2(float phi)
{
	// Solve: '2.theta + sin(2.theta) - 2.sin(theta) = k.sin(phi)' w.r.t 'theta'
	const float phi2 = rsqr(phi);
	const float theta = phi * 0.615709F + phi2 * (0.00909953F + phi2 * 0.0046292F);
	return SolveTwoThetaPlusSinTwoThetaMinusTwoSinTheta(1.2283697F * rsin(phi), theta);
}

local float SolvePutnins6(float a, float b, float phi)
{
	// Solve the Putnins P6 and P6' equations
	const float rhs = b * rsin(phi);
	float theta = phi * 1.10265779F;
	int i;
	if (rabs(phi) > 1.5707963F)
		return rsgn(phi) * 1.732050807568877F;
	for (i = 0; i < 8; i++)
	{
		const float r = rsqrt(1 + rsqr(theta));
		const float p = ((a - r) * theta) - rln(theta + r) - rhs;
		const float q = a - 2 * r;
		ASSERT(q > 0);
		theta -= p / q;
		if (rabs(p) < 0.000001F)
			break;
	}
	return theta;
}

local float SolveHatano(float b, float phi)
{
	// Solve: '2.theta + sin(2.theta) = b.sin(phi)' w.r.t 'theta'
	return SolveThetaPlusSinTheta(b * rsin(phi), phi) / 2;
}

local void ProjectRectangular(const MAPPROJECT* p)
{
	// Just the x-axis
	const int rx = ((p->rx + p->scroll) % 24) * 15;
	int i;
	for (i = 0; i < p->n; i++)
		p->dst[i].x = rx + p->src[i].x;
	return;
}

local int PROJAPI ProjectMercator(const MAPPROJECT* p)
{
	// Mercator projection (p.46)
	const int ry = 90 - p->ry * 30;
	int i;
	for (i = 0; i < p->n; i++)
	{
		// Use Mercator's own range of latitudes
		float lat = rclamp(ry - p->src[i].y, -74.0F, 83.3F);
		p->dst[i].y = 106.5F - rln(rtan(deg2rad(45 + lat / 2))) * 37.6F;
	}
	ProjectRectangular(p);
	return p->n;
}

local int PROJAPI ProjectTrapizoidal(const MAPPROJECT* p)
{
	// Trapizoidal (p.8)
	const float scale = p->projection->parameter / 360.0F;
	const int rx = ((p->rx + p->scroll) % 24) * 15;
	const int ry = p->ry * 30;
	int i;
	for (i = 0; i < p->n; i++)
	{
		const float lat = ry + p->src[i].y;
		float a = 1 - rabs(lat / 90 - 1) * scale;
		p->dst[i].x = (rx + p->src[i].x - 180) * a + 180;
		p->dst[i].y = lat;
	}
	return p->n;
}

local int PROJAPI ProjectPtolemy1(const MAPPROJECT* p)
{
	// Ptolemy's first (p.11)
	const float scale = 1.54F;
	const float offset = -23.0F;
	const int rx = ((p->rx + p->scroll + 20) % 24 - 12) * 15;
	int i;
	if ((rx < -90) || (rx >= 90) || (p->ry > 3))
		return 0;
	if (p->ry < 3)
	{
		// Northern hemisphere
		const int ry = p->ry * 30 + 25;
		for (i = 0; i < p->n; i++)
		{
			const float theta = deg2rad(((rx + p->src[i].x) * 105) / 180);
			const float r = (ry + p->src[i].y) * scale;
			p->dst[i].x = 180 + r * rsin(theta);
			p->dst[i].y = r * rcos(theta) + offset;
		}
	}
	else
	{
		// Southern hemisphere
		static const float max_s = dms2deg(16, 25, 0);
		for (i = 0; i < p->n; i++)
		{
			float r = rmin(p->src[i].y, max_s);
			const float theta = deg2rad(((rx + p->src[i].x) * (105 - r * 2)) / 180);
			r = (r + 115) * scale;
			p->dst[i].x = 180 + r * rsin(theta);
			p->dst[i].y = r * rcos(theta) + offset;
		}
	}
	return p->n;
}

local int PROJAPI ProjectPtolemy2(const MAPPROJECT* p)
{
	// Ptolemy's second (p.12)
	const float min_s = dms2deg(27, 0, 0);
	const float max_s = dms2deg(106, 25, 0);
	int i, ry, rx = ((p->rx + p->scroll + 20) % 24 - 12) * 15;
	if ((rx < -90) || (rx >= 90) || (p->ry > 3))
		return 0;
	ry = p->ry * 30;
	for (i = 0; i < p->n; i++)
	{
		float r = rclamp(ry + p->src[i].y, min_s, max_s);
		float a = rsqrt(1 - rsqr((r + 23) / 90 - 1));
		const float theta = deg2rad((rx + p->src[i].x) / 6) * (a * 4 - 2);
		r = (r + 90) * 2;
		p->dst[i].x = 180 + r * rsin(theta);
		p->dst[i].y = r * rcos(theta) - 213;
	}
	return p->n;
}

local int PROJAPI ProjectGlobular(const MAPPROJECT* p)
{
	// Bacon's globular (p.15)
	const int rx = ((p->rx + p->scroll + 18) % 24) * 15;
	const int ry = p->ry * 30;
	const int cx = (rx < 180) ? 90 : 270;
	const int dx = (rx < cx) ? 1 : -1;
	int i;
	for (i = 0; i < p->n; i++)
	{
		float x = dx * (rx - cx + p->src[i].x) / 90;
		const float y = rcos(deg2rad(ry + p->src[i].y));
		if (x < -0.000001F)
		{
			const float a = (1 - rsqr(x)) / (x * 2);
			x = dx * (rsqrts(rsqr(a) - rsqr(y) + 1) + a);
		}
		p->dst[i].x = cx - x * 90;
		p->dst[i].y = 90 - y * 90;
	}
	return p->n;
}

local int PROJAPI ProjectEquatorialOrthographic(const MAPPROJECT* p)
{
	// Equatorial orthographic (p.16)
	const int rx = ((p->rx + p->scroll + 18) % 24) * 15;
	const int ry = p->ry * 30;
	int i, dx = (rx >= 180) ? 1 : -1;
	for (i = 0; i < p->n; i++)
	{
		const float lon = deg2rad(rx + p->src[i].x);
		const float lat = deg2rad(ry + p->src[i].y);
		p->dst[i].x = 180 + 90 * (dx + rsin(lat) * rcos(lon) * dx);
		p->dst[i].y = 90 - 90 * rcos(lat);
	}
	return p->n;
}

local int PROJAPI ProjectPolarOrthographic(const MAPPROJECT* p)
{
	// Polar orthographic (p.16)
	const int rx = ((p->rx + p->scroll) % 24) * 15;
	const int ry = p->ry * 30;
	int i, dx = (ry >= 90) ? 1 : -1;
	for (i = 0; i < p->n; i++)
	{
		const float lon = deg2rad(rx + p->src[i].x);
		const float lat = deg2rad(ry + p->src[i].y);
		const float r = rsin(lat);
		p->dst[i].x = 180 + 90 * (dx - r * rsin(lon));
		p->dst[i].y = 90 + 90 * dx * r * rcos(lon);
	}
	return p->n;
}

local int PROJAPI ProjectObliqueOrthographic(const MAPPROJECT* p)
{
	// Oblique orthographic (p.17)
	static const BYTE skip[] = { 0, 1, 4, 6, 7 };
	int i = (p->rx + p->scroll + 12) % 24;
	const int rx = ((i + 6) % 24) * 15;
	const int ry = p->ry * 30;
	if (p->ry >= dimof(skip))
		return 0;
	if ((i >= (12 - skip[p->ry])) && (i < (12 + skip[p->ry])))
		return 0;
	for (i = 0; i < p->n; i++)
	{
		// Tilt forwards by 30 degrees
		const float lon = deg2rad(rx + p->src[i].x);
		const float lat = deg2rad(ry + p->src[i].y);
		float x = -rcos(lon) * rsin(lat);
		const float ay = rcos(lat);
		const float az = -rsin(lon) * rsin(lat);
		float y = ay * cos_30 + az * sin_30;
		const float z = az * cos_30 - ay * sin_30;
		if (z > 0)
		{
			// Place on the rim
			const float r = rsqrt(1 - z * z);
			ASSERT(r > 0);
			x /= r;
			y /= r;
		}
		p->dst[i].x = 180 + 90 * x;
		p->dst[i].y = 90 - 90 * y;
	}
	return p->n;
}

local int PROJAPI ProjectEquatorialGnomonic(const MAPPROJECT* p)
{
	// Equatorial gnomonic (p.20)
	const int rx = ((p->rx + p->scroll + 18) % 24) * 15;
	const int ry = p->ry * 30;
	int i;
	if ((rx == 0) || (rx >= 165) || (ry == 0) || (ry == 150))
		return 0;
	for (i = 0; i < p->n; i++)
	{
		const float lon = deg2rad(rx + p->src[i].x - 90);
		const float lat = deg2rad(ry + p->src[i].y - 90);
		p->dst[i].x = 180 + 60 * rtan(lon);
		p->dst[i].y = 90 + 60 * rtan(lat) / rcos(lon);
	}
	return p->n;
}

local int PROJAPI ProjectPolarGnomonic(const MAPPROJECT* p)
{
	// Polar gnomonic (p.18)
	const int rx = ((p->rx + p->scroll + 18) % 24) * 15;
	const int ry = p->ry * 30;
	int i;
	if (ry >= 90)
		return 0;
	for (i = 0; i < p->n; i++)
	{
		const float lon = deg2rad(rx + p->src[i].x);
		const float lat = deg2rad(rmin(ry + p->src[i].y, 80));
		const float r = 50 * rtan(lat);
		p->dst[i].x = 180 - r * rcos(lon);
		p->dst[i].y = 90 + r * rsin(lon);
	}
	return p->n;
}

local int PROJAPI ProjectObliqueGnomonic(const MAPPROJECT* p)
{
	// Oblique gnomonic at 45 degrees (p.19,89)
	const int rx = ((p->rx + p->scroll) % 24) * 15 - 180;
	const int ry = p->ry * 30 - 90;
	float r1, r2;
	int i;
	if (ry > 0)
		return 0;
	for (i = 0; i < p->n; i++)
	{
		const float lon = deg2rad(rx + p->src[i].x);
		const float lat = deg2rad(ry + p->src[i].y);
		float cos_z = cos_60 * rcos(lat) * rcos(lon) - sin_60 * rsin(lat);
		if (cos_z <= 0)
		{
			// Peg the latitude
			r1 = 250;
			r2 = 250;
		}
		else
		{
			// Normal gnomonic
			r1 = rmin(35 / cos_z, 250);
			r2 = rcos(lat) * r1;
		}
		p->dst[i].x = 180 + r2 * rsin(lon);
		p->dst[i].y = 90 + r2 * rcos(lon) + r1 * cos_60 * rsin(lat);
	}
	return p->n;
}

local int PROJAPI ProjectEquatorialStereographic(const MAPPROJECT* p)
{
	// Equatorial stereographic (p.28)
	const int rx = ((p->rx + p->scroll + 18) % 24) * 15;
	const int ry = p->ry * 30 - 90;
	int i, dx = (rx >= 180) ? -1 : 1;
	for (i = 0; i < p->n; i++)
	{
		const float lon = dx * deg2rad(rx + p->src[i].x);
		const float lat = deg2rad(ry + p->src[i].y);
		float r = 1 + rsin(lon) * rcos(lat);
		ASSERT(r != 0);
		r = 1 / r;
		ASSERT((r >= 0) && (r < 1.000001F));
		p->dst[i].x = 180 - 90 * (rcos(lat) * rcos(lon) * r * dx + dx);
		p->dst[i].y = 90 + 90 * rsin(lat) * r;
	}
	return p->n;
}

local int PROJAPI ProjectObliqueStereographic(const MAPPROJECT* p)
{
	// Oblique stereographic at 30 degrees (p.24)
	static const BYTE skip[] = { 0, 0, 4, 6, 7 };
	int i = (p->rx + p->scroll + 12) % 24;
	const int rx = i * 15;
	const int ry = p->ry * 30 - 90;
	if (p->ry >= dimof(skip))
		return 0;
	if ((i >= (12 - skip[p->ry])) && (i < (12 + skip[p->ry])))
		return 0;
	for (i = 0; i < p->n; i++)
	{
		const float lon = deg2rad(rx + p->src[i].x);
		const float lat = deg2rad(ry + p->src[i].y);
		float x, y, r = 1 - sin_30 * rsin(lat) + cos_30 * rcos(lat) * rcos(lon);
		ASSERT(r > 0);
		x = rcos(lat) * rsin(lon) / r;
		y = (cos_30 * rsin(lat) + sin_30 * rcos(lat) * rcos(lon)) / r;
		r = rsqr(x) + rsqr(y);
		if (r > 1)
		{
			// Fit to circle
			r = rsqrt(r);
			ASSERT(r > 0);
			x /= r;
			y /= r;
		}
		p->dst[i].x = 180 + 90 * x;
		p->dst[i].y = 90 + 90 * y;
	}
	return p->n;
}

local int PROJAPI ProjectAzimuthalEquidistant(const MAPPROJECT* p)
{
	// Azimuthal equidistant (p.29)
	const int rx = ((p->rx + p->scroll + 12) % 24) * 15;
	const int ry = p->ry * 30 - 90;
	int i, dx = (ry >= 0) ? -1 : 1;
	for (i = 0; i < p->n; i++)
	{
		const float lon = deg2rad(rx + p->src[i].x);
		const float lat = 90 + dx * (ry + p->src[i].y);
		p->dst[i].x = 180 + lat * rsin(lon) - 90 * dx;
		p->dst[i].y = 90 + lat * rcos(lon) * dx;
	}
	return p->n;
}

local int PROJAPI ProjectAzimuthalEquidistantInterrupted(const MAPPROJECT* p)
{
	// Azimuthal equidistant, interrupted (p.29)
	const int rx = ((p->rx + p->scroll) % 24) * 15;
	const int ry = p->ry * 30 - 90;
	int i, dx = (ry >= 0) ? -1 : 1;
	int cx = (ry < 0) ? 180 : (rx < 180) ? 0 : 360;
	for (i = 0; i < p->n; i++)
	{
		const float lon = deg2rad(rx + p->src[i].x + cx);
		const float lat = 90 + dx * (ry + p->src[i].y);
		p->dst[i].x = cx + lat * rsin(lon);
		p->dst[i].y = 90 + lat * rcos(lon) * dx;
	}
	return p->n;
}

local int PROJAPI ProjectAzimuthalEquidistantWorldwide(const MAPPROJECT* p)
{
	// Azimuthal equidistant, worldwide (p.29)
	const int rx = ((p->rx + p->scroll + 12) % 24) * 15;
	const int ry = p->ry * 30;
	int i;
	for (i = 0; i < p->n; i++)
	{
		const float lon = deg2rad(rx + p->src[i].x);
		const float lat = (ry + p->src[i].y) / 2;
		p->dst[i].x = 180 + lat * rsin(lon);
		p->dst[i].y = 90 + lat * rcos(lon);
	}
	return p->n;
}

local int PROJAPI ProjectAzimuthalEquidistantEquatorial(const MAPPROJECT* p)
{
	// Equatorial azimuthal equidistant: See Aitoff (p.131)
	const int rx = ((p->rx + p->scroll + 18) % 24) * 15;
	const int ry = p->ry * 30 - 90;
	const int cx = (rx < 180) ? 90 : 270;
	int i;
	for (i = 0; i < p->n; i++)
	{
		const float lon = deg2rad(rx + p->src[i].x - cx);
		const float lat = deg2rad(ry + p->src[i].y);
		const float z = racos(rcos(lat) * rcos(lon));
		const float r = (z < 0.000001F) ? 0 : z * (180 / pi) / rsin(z);
		p->dst[i].x = cx + r * rcos(lat) * rsin(lon);
		p->dst[i].y = 90 + r * rsin(lat);
	}
	return p->n;
}

local int PROJAPI ProjectSimpleConic(const MAPPROJECT* p)
{
	// Simple conic (p.29)
	const int rx = ((p->rx + p->scroll) % 24 - 12) * 15;
	const int ry = p->ry * 30 + 25;
	int i;
	if (ry > 115)
		return 0;
	for (i = 0; i < p->n; i++)
	{
		const float theta = deg2rad(((rx + p->src[i].x) * 105) / 180);
		const float r = rmin((ry + p->src[i].y), 125) * 1.14F;
		p->dst[i].x = 180 + r * rsin(theta);
		p->dst[i].y = r * rcos(theta) + 37.5F;
	}
	return p->n;
}

local int PROJAPI ProjectEquidistantConic(const MAPPROJECT* p)
{
	// Equidistant conic (p.31)
	const int rx = ((p->rx + p->scroll) % 24 - 12) * 15;
	const int ry = p->ry * 30;
	int i;
	if (ry > 120)
		return 0;
	for (i = 0; i < p->n; i++)
	{
		const float theta = deg2rad(((rx + p->src[i].x) * 105) / 180);
		const float r = rmin((ry + p->src[i].y), 128) * 1.11F;
		p->dst[i].x = 180 + r * rsin(theta);
		p->dst[i].y = r * rcos(theta) + 37.5F;
	}
	return p->n;
}

local int PROJAPI ProjectBonne(const MAPPROJECT* p)
{
	// Bonne at 47 degrees (p.34)
	const int rx = ((p->rx + p->scroll) % 24 - 12) * 15;
	const int ry = p->ry * 30 - 90;
	int i;
	if ((rx < -150) || (rx >= 150) || (ry > 30))
		return 0;
	for (i = 0; i < p->n; i++)
	{
		const float lat = deg2rad(rclamp(ry + p->src[i].y, -80, 40));
		float r = 1.75282F + lat;
		float theta = rcos(lat) * deg2rad(rx + p->src[i].x) / r;
		r *= 64;
		p->dst[i].x = 180 + r * rsin(theta);
		p->dst[i].y = 23 + r * rcos(theta);
	}
	return p->n;
}

local int PROJAPI ProjectWerner(const MAPPROJECT* p)
{
	// Werner 1 (p.35)
	const int werner = p->projection->parameter;
	const int rx = ((p->rx + p->scroll) % 24 - 12) * 15;
	const int ry = p->ry * 30 - 90;
	const int cy = (werner == 1) ? 25 : 41;
	float r = 0, theta = 0;
	int i;
	if ((werner == 1) && ((rx < -90) || (rx >= 90)))
		return 0;
	for (i = 0; i < p->n; i++)
	{
		const float lon = deg2rad(rx + p->src[i].x);
		const float lat = deg2rad(ry + p->src[i].y);
		switch (werner)
		{
			case 1:
				// Werner 1
				r = 3.14159F + lat * 2;
				theta = rcos(lat) * lon * pi / r;
				r *= 24.6F;
				break;
			case 2:
				// Werner 2:
				r = 1.5708F + lat;
				theta = rcos(lat) * lon / r;
				r *= 43;
				break;
			case 3:
				// Werner 3:
				r = 1.5471975512F + lat;
				theta = (rabs(r) < 0.000001F) ? 0 : rclamp(rcos(lat) * lon / r, -pi, pi);
				r *= 43;
				break;
			default:
				UNREACHABLE();
		}
		p->dst[i].x = 180 + r * rsin(theta);
		p->dst[i].y = cy + r * rcos(theta);
	}
	return p->n;
}

local int PROJAPI ProjectDoubleCordiform(const MAPPROJECT* p)
{
	// Double cordiform (p.37)
	const int rx = ((p->rx + p->scroll) % 24 - 12) * 15;
	const int ry = p->ry * 30 - 90;
	const int dx = (ry < 0) ? -1 : 1;
	int i;
	for (i = 0; i < p->n; i++)
	{
		const float lat = deg2rad(ry + p->src[i].y);
		float r = 1.5708F - rabs(lat);
		float theta = rcos(lat) * deg2rad(rx + p->src[i].x) / r;
		r *= 57.2F;
		p->dst[i].x = 180 + dx * (90 - r * rcos(theta));
		p->dst[i].y = 90 - r * rsin(theta);
	}
	return p->n;
}

local int PROJAPI ProjectFourLobedStar(const MAPPROJECT* p)
{
	// Four-lobed star (p.38)
	const int rx = ((p->rx + p->scroll + 6) % 24) * 15;
	const int ry = p->ry * 30;
	const int lobe = rx / 90;
	const float spike = (lobe * 2 + 1) * pi / 4;
	int i;
	for (i = 0; i < p->n; i++)
	{
		const float lon = deg2rad(rx + p->src[i].x);
		const float lat = deg2rad(ry + p->src[i].y);
		const float r = (ry + p->src[i].y) * sin_45;
		float theta = lon + (spike - lon) * lat / pi;
		p->dst[i].x = 180 + r * rcos(theta);
		p->dst[i].y = 90 - r * rsin(theta);
	}
	return p->n;
}

local int PROJAPI ProjectBordone(const MAPPROJECT* p)
{
	// Bordone (p.38)
	const int rx = ((p->rx + p->scroll) % 24 - 12) * 15;
	const int ry = p->ry * 30 - 90;
	int i;
	for (i = 0; i < p->n; i++)
	{
		const float lon = rx + p->src[i].x;
		const float lat = ry + p->src[i].y;
		p->dst[i].x = 180 + rsqrts(1 - rsqr(lat / 90)) * lon;
		p->dst[i].y = 90 + lat;
	}
	return p->n;
}

local int PROJAPI ProjectOval(const MAPPROJECT* p)
{
	// Oval (p.39)
	const int rx = ((p->rx + p->scroll) % 24 - 12) * 15;
	const int ry = p->ry * 30 - 90;
	const int dx = (rx < 0) ? 1 : -1;
	int i;
	if ((rx >= -90) && (rx < 90))
	{
		// Central globular
		for (i = 0; i < p->n; i++)
		{
			float x = dx * (rx + p->src[i].x) / 90;
			const float y = (ry + p->src[i].y) / 90;
			if (x > -0.000001F)
				x = 0;
			else
			{
				const float a = (1 - rsqr(x)) / (x * 2);
				x = dx * (rsqrts(rsqr(a) - rsqr(y) + 1) + a);
			}
			p->dst[i].x = 180 - x * 90;
			p->dst[i].y = 90 + y * 90;
		}
	}
	else
	{
		// Outer sections
		for (i = 0; i < p->n; i++)
		{
			const float y = ry + p->src[i].y;
			p->dst[i].x = 180 - dx * (rsqrts(8100 - rsqr(y)) - 90) + rx + p->src[i].x;
			p->dst[i].y = 90 + y;
		}
	}
	return p->n;
}

local int PROJAPI ProjectApian1(const MAPPROJECT* p)
{
	// Apian I (p.14)
	const int rx = ((p->rx + p->scroll + 18) % 24) * 15;
	const int ry = p->ry * 30 - 90;
	const int cx = (rx < 180) ? 90 : 270;
	const int dx = (rx < cx) ? 1 : -1;
	int i;
	for (i = 0; i < p->n; i++)
	{
		float x = dx * (rx - cx + p->src[i].x) / 90;
		const float y = (ry + p->src[i].y) / 90;
		if (x < -0.000001F)
		{
			const float a = (1 - rsqr(x)) / (x * 2);
			x = dx * (rsqrts(rsqr(a) - rsqr(y) + 1) + a);
		}
		p->dst[i].x = cx - x * 90;
		p->dst[i].y = 90 + y * 90;
	}
	return p->n;
}

local int PROJAPI ProjectApian2(const MAPPROJECT* p)
{
	// Apian II (p.14)
	const int rx = ((p->rx + p->scroll + 18) % 24) * 15;
	const int ry = p->ry * 30 - 90;
	const int cx = (rx < 180) ? 90 : 270;
	int i;
	for (i = 0; i < p->n; i++)
	{
		const float y = (ry + p->src[i].y) / 90;
		const float x = rsqrts(1 - rsqr(y)) * (rx - cx + p->src[i].x);
		p->dst[i].x = cx + x;
		p->dst[i].y = 90 + y * 90;
	}
	return p->n;
}

local int PROJAPI ProjectOctant(const MAPPROJECT* p)
{
	// Octant (p.40)
	static const BYTE cxi[4] = { 93, 93, 87, 87 };
	static const BYTE cyi[4] = { 93, 87, 87, 93 };
	int i = (p->rx + p->scroll + 12) % 24;
	const int ix = i / 6;
	const int iy = p->ry / 3;
	const int rx = i * 15 - ix * 90 - 45;
	const int ry = p->ry * 30;
	const int cx = cxi[ix] + 180 * iy;
	const int dx = 1 - iy * 2;
	const int cy = (cyi[ix] - 90) * dx + 90;
	const float cr = (ix * 2 + 1) * pi / 4;
	for (i = 0; i < p->n; i++)
	{
		const float r = (iy == 0) ? (ry + p->src[i].y) : (180 - ry - p->src[i].y);
		const float theta = (racos(r / 180) - pi / 6) * (rx + p->src[i].x) / 45 + cr;
		p->dst[i].x = cx + r * rsin(theta);
		p->dst[i].y = cy + r * rcos(theta) * dx;
	}
	return p->n;
}

local int PROJAPI ProjectAitoff(const MAPPROJECT* p)
{
	// Aitoff (p.131)
	const int hammer = p->projection->parameter;
	const int rx = ((p->rx + p->scroll) % 24 - 12) * 15;
	const int ry = p->ry * 30 - 90;
	float r;
	int i;
	for (i = 0; i < p->n; i++)
	{
		const float lon = deg2rad(rx + p->src[i].x) / 2;
		const float lat = deg2rad(ry + p->src[i].y);
		if (hammer)
		{
			// Hammer (p.133)
			r = 90 / rsqrt(1 + rcos(lat) * rcos(lon));
		}
		else
		{
			const float z = racos(rcos(lat) * rcos(lon));
			r = (z < 0.000001F) ? 0 : z * (180 / pi) / rsin(z);
		}
		p->dst[i].x = 180 + 2 * r * rcos(lat) * rsin(lon);
		p->dst[i].y = 90 + r * rsin(lat);
	}
	return p->n;
}

local int PROJAPI ProjectNicolosi(const MAPPROJECT* p)
{
	// Nicolosi globular (p.41)
	const int gore = p->projection->parameter;
	const int rx = ((p->rx + p->scroll + 18) % 24) * 15;
	const int ry = 90 - p->ry * 30;
	const int cx = (rx / gore) * gore + (gore / 2);
	float x, y;
	int i;
	for (i = 0; i < p->n; i++)
	{
		const float lon = deg2rad(rx + p->src[i].x - cx);
		const float lat = deg2rad(ry - p->src[i].y);
		if (rabs(lon) < 0.000001F)
		{
			// Central meridian
			x = 0;
			y = lat;
		}
		else if (rabs(lat) < 0.000001F)
		{
			// Equator
			x = lon;
			y = 0;
		}
		else if (rabs(lon) > 1.57F)
		{
			// Edge
			x = lon * rcos(lat);
			y = pi / 2 * rsin(lat);
		}
		else if (rabs(lat) > 1.57F)
		{
			// Edge
			x = 0;
			y = lat;
		}
		else
		{
			// Inside the circle
			const float tb = pi / (lon * 2) - (lon * 2) / pi;
			const float c = lat * 2 / pi;
			const float sp = rsin(lat);
			const float d = (1 - rsqr(c)) / (sp - c);
			const float r2 = rsqr(tb / d);
			const float m = (tb * sp / d - tb / 2) / (1 + r2);
			const float n = (sp / r2 + d / 2) / (1 + 1 / r2);
			x = rsqr(m) + rsqr(rcos(lat)) / (1 + r2);
			x = (m + rsgn(lon) * rsqrts(x)) * pi / 2;
			y = rsqr(n) - (rsqr(sp) / r2 + d * sp - 1) / (1 + 1 / r2);
			y = (n - rsgn(lat) * rsqrts(y)) * pi / 2;
		}
		p->dst[i].x = cx + x / pi * 180;
		p->dst[i].y = 90 - y / pi * 180;
	}
	return p->n;
}

local int PROJAPI ProjectPolarFar(const MAPPROJECT* p)
{
	// Polar stereographic, La Hire, Parent, etc (p.22-65)
	static const float distance[] = { 1, 1.707106781F, 1.594F, 1.732050808F, 2.105F };
	const float d = distance[p->projection->parameter];
	const int rx = ((p->rx + p->scroll) % 24) * 15 - 180;
	const int ry = p->ry * 30 - 90;
	const int dx = (ry < 0) ? 1 : -1;
	const int cx = 180 - 90 * dx;
	int i;
	ASSERT((p->projection->parameter >= 0) && (p->projection->parameter < dimof(distance)));
	for (i = 0; i < p->n; i++)
	{
		const float lon = deg2rad(rx + p->src[i].x);
		const float lat = deg2rad(90 - rabs(ry + p->src[i].y));
		const float r = rsin(lat) / (d + rcos(lat)) * d * 90;
		p->dst[i].x = cx + r * rsin(lon);
		p->dst[i].y = 90 + r * rcos(lon) * dx;
	}
	return p->n;
}

local int PROJAPI ProjectConic(const MAPPROJECT* p)
{
	// Conic with two parallels, De l'Isle, Murdoch, Euler, etc (pp.68)
	static const float na[][2] =
	{
		// Conic constants (assume limiting/standard latitudes are 30 and 60 degrees north)
		{ 0.6366197723675813430755350534901F, 1.7334576123659682453004763924874F },
		{ sin_45, 1.6857144795545543791708600368266F },
		{ 0.5946035575013605333587499852802F, 0.8408964152537145430311254762332F },
		{ 0.8105694691387021715510357056778F, pi / 2 },
		{ sin_45, 0.9659258262890682867497431997289F },
		{ sin_45, 1.6817928305074290860622509524664F },
		{ sin_45, pi / 2 },
		{ 0.7155668471806278662776917314954F, 0 },
		{ 0.6830127018922193233818615853765F, 1.433012701892219323381861585376F }
	};
	const int conic = p->projection->parameter;
	const int rx = ((p->rx + p->scroll) % 24 - 12) * 15;
	const int ry = 90 - p->ry * 30;
	float a, n, r = 0;
	int i;
	if ((ry <= 0) || (rx < -90) || (rx >= 90))
		return 0;
	ASSERT((conic >= 0) && (conic <= dimof(na)));
	n = na[conic][0];
	a = na[conic][1];
	for (i = 0; i < p->n; i++)
	{
		const float lon = deg2rad(rx + p->src[i].x);
		const float lat = deg2rad(ry - p->src[i].y);
		const float theta = lon * n;
		switch (conic)
		{
			case 2:
				// Murdoch II
				r = rmax(a + rtan(pi / 4 - lat), 0) * 180 / (a + 1);
				break;
			case 4:
				// Perspective
				r = (a * (1 - rtan(lat - pi / 4))) * 183 / (a + 1);
				break;
			case 5:
				// Tissot
				r = rsqrts(2 * (1 - rsin(lat)) / n) * 180 / a;
				break;
			case 7:
				// Lambert I
				r = 180 / rpow(rtan(pi / 4 + rmin(lat, 1.5707963F) / 2), n);
				break;
			case 8:
				// Albers
				r = 102.7F * rsqrt(a - 2 * n * rsin(lat)) / n;
				break;
			default:
				// Euler, Murdoch I & III, Vitkovskiy II
				r = (a - lat) * 180 / a;
				break;
		}
		ASSERT(r > -0.00001F);
		p->dst[i].x = 180 + r * rsin(theta);
		p->dst[i].y = r * rcos(theta);
	}
	return p->n;
}

local int PROJAPI ProjectTransverse(const MAPPROJECT* p)
{
	// Cassini and Lambert (p.75,84,87)
	const int transverse = p->projection->parameter;
	const int rx = ((p->rx + p->scroll + 13) % 24) * 15;
	const int ry = 90 - p->ry * 30;
	const int cx = ((rx < 90) || (rx >= 270)) ? 180 : (ry <= 0) ? 360 : 0;
	float y = 0;
	int i;
	for (i = 0; i < p->n; i++)
	{
		const float lon = deg2rad(rx + p->src[i].x);
		const float lat = deg2rad(ry - p->src[i].y);
		const float t = rclamp(lat, -1.5707963F, 1.5707963F);
		p->dst[i].x = (rabs(lat) > 0.000001F) ? 180 - ratan(rtan(t), rcos(lon)) / pi * 180 : cx;
		switch (transverse)
		{
			case 0:
				// Cassini
				y = rasin(rcos(lat) * rsin(lon)) / pi * 183;
				break;
			case 1:
				// Lambert III
				y = rclamp(rcos(lat) * rsin(lon), -0.7616F, 0.7616F);
				y = 45 * rln((1 + y) / (1 - y));
				break;
			case 2:
				// Lambert V
				y = 90 * rcos(lat) * rsin(lon);
				break;
			case 3:
				// Wetch
				y = 1 - rsqr(rcos(t)) * rsqr(rsin(lon));
				if (y > 0.000000000001F)
					y = 90 * rcos(t) * rsin(lon) / rsqrt(y);
				else
					y = (float)((rx < 180) ? 91 : -91);
				break;
			default:
				UNREACHABLE();
		}
		p->dst[i].y = rclamp(90 - y, -1, 181);
	}
	return p->n;
}

local int PROJAPI ProjectLagrange(const MAPPROJECT* p)
{
	// Lagrange (p.80)
	const int rx = ((p->rx + p->scroll + 1) % 24 - 12) * 15;
	const int ry = 90 - p->ry * 30;
	int i;
	for (i = 0; i < p->n; i++)
	{
		const float lon = deg2rad(rx + p->src[i].x) / 2;
		const float lat = deg2rad(ry - p->src[i].y) / 2;
		const float sin_phi = rtan(lat);
		const float cos_phi = rsqrts(1 - rsqr(sin_phi));
		const float a = 1 + cos_phi * rcos(lon);
		if (rabs(a) > 0.000001F)
		{
			// Main section
			p->dst[i].x = 180 + 90 * cos_phi * rsin(lon) / a;
			p->dst[i].y = 90 - 90 * sin_phi / a;
		}
		else
		{
			// Poles
			p->dst[i].x = 180;
			p->dst[i].y = 90 - lat / pi * 360;
		}
	}
	return p->n;
}

local int PROJAPI ProjectCylindricalEqualArea(const MAPPROJECT* p)
{
	// Lambert IV (p.87)
	const int ry = 90 - p->ry * 30;
	int i;
	for (i = 0; i < p->n; i++)
		p->dst[i].y = 90 - 90 * rsin(deg2rad(ry - p->src[i].y));
	ProjectRectangular(p);
	return p->n;
}

local int PROJAPI ProjectAzimuthalPolar(const MAPPROJECT* p)
{
	// Azimuthal polar (p.89)
	const int rx = ((p->rx + p->scroll) % 24) * 15;
	const int ry = 90 - p->ry * 30;
	const int dx = (ry > 0) ? 1 : -1;
	int i;
	for (i = 0; i < p->n; i++)
	{
		float theta = deg2rad(rx + p->src[i].x);
		float zz, z = deg2rad(90 - dx * (ry - p->src[i].y));
		float r = rsin(z / 2) / sin_45 * 90;
		switch (p->projection->parameter)
		{
			case 0:
				// Lambert (p.89)
				break;
			case 1:
				// Wiechel (p.131)
				theta -= z / 2;
				break;
			case 2:
				// Ginzburg I (p.225)
				zz = z * z * z;
				r = (r + 0.042F * zz * zz * zz) * 0.9735489F;
				break;
			case 3:
				// Ginzburg II (p.225)
				zz = z * z * z;
				r = (r + 0.01591F * zz * zz * zz * z) * 0.98409F;
				break;
			default:
				UNREACHABLE();
		}
		p->dst[i].x = 180 - r * rsin(theta) - 90 * dx;
		p->dst[i].y = 90 - r * rcos(theta) * dx;
	}
	return p->n;
}

local int PROJAPI ProjectLambertEquatorial(const MAPPROJECT* p)
{
	// Lambert azimuthal equatorial (p.89)
	const int rx = ((p->rx + p->scroll + 18) % 24) * 15;
	const int ry = 90 - p->ry * 30;
	const int cx = (rx < 180) ? 90 : 270;
	int i;
	for (i = 0; i < p->n; i++)
	{
		const float lon = deg2rad(rx + p->src[i].x - cx);
		const float lat = deg2rad(ry - p->src[i].y);
		float r = 1 + rcos(lat) * rcos(lon);
		ASSERT(r > 0.000001F);
		r = rsqrt(1 / r) * 90;
		p->dst[i].x = cx + r * rcos(lat) * rsin(lon);
		p->dst[i].y = 90 - r * rsin(lat);
	}
	return p->n;
}

local int PROJAPI ProjectSemicircleConical(const MAPPROJECT* p)
{
	// Lambert and Braun conical (p.91,124)
	const int braun = p->projection->parameter;
	const int rx = ((p->rx + p->scroll) % 24) * 15;
	const int ry = 90 - p->ry * 30;
	int i;
	for (i = 0; i < p->n; i++)
	{
		const float lat = deg2rad(ry - p->src[i].y);
		const float theta = deg2rad(rx + p->src[i].x) / 2;
		const float r = braun ?	(cos_30 - rtan((lat + pi / 6) / 2) / 2) * 155.88F
							  :	rsin(pi / 4 - lat / 2) * 180;
		p->dst[i].x = 180 - r * rcos(theta);
		p->dst[i].y = r * rsin(theta);
	}
	return p->n;
}

local int PROJAPI ProjectCentralCylindrical(const MAPPROJECT* p)
{
	// Central cylindrical (p.107)
	const int ry = p->ry * 30 - 90;
	int i;
	for (i = 0; i < p->n; i++)
	{
		// Clamp to non-polar regions
		const float lat = rclamp(ry + p->src[i].y, -66.33F, 66.33F);
		p->dst[i].y = 90 + rtan(deg2rad(lat)) * 39.451F;
	}
	ProjectRectangular(p);
	return p->n;
}

local int PROJAPI ProjectObliqueMercator(const MAPPROJECT* p)
{
	// Oblique Mercator (p.107)
	const int rx = ((p->rx + p->scroll) % 24 - 12) * 15;
	const int ry = 90 - p->ry * 30;
	int i;
	for (i = 0; i < p->n; i++)
	{
		// Make oblique before performing the standard Mercator
		const float x = deg2rad(rx + p->src[i].x);
		const float y = deg2rad(ry - p->src[i].y);
		const float lat = rasin(sin_45 * rsin(y) - cos_45 * rcos(y) * rcos(x));
		float lon = ratan(rcos(y) * rsin(x), sin_45 * rcos(y) * rcos(x) + cos_45 * rsin(y));
		if (((rx + 1) * lon) < 0)
			if (rx >= 0)
				lon = (lon < -1) ? pi : 0;
			else
				lon = (lon > 1) ? -pi : 0;
		p->dst[i].x = 180 + lon / pi * 180;
		p->dst[i].y = 90 - rtan(rclamp(lat, -1, 1)) * 57.8F;
	}
	return p->n;
}

local int PROJAPI ProjectGall(const MAPPROJECT* p)
{
	// Gall stereographic (p.110)
	const int ry = p->ry * 30 - 90;
	int i;
	for (i = 0; i < p->n; i++)
		p->dst[i].y = 90 + 90 * rtan(deg2rad(ry + p->src[i].y) / 2);
	ProjectRectangular(p);
	return p->n;
}

local int PROJAPI ProjectBraun(const MAPPROJECT* p)
{
	// Braun I (p.110)
	const int ry = 90 - p->ry * 30;
	int i;
	for (i = 0; i < p->n; i++)
	{
		// Clamp to interesting regions
		const float lat = deg2rad(rclamp(ry - p->src[i].y, -85, 85));
		p->dst[i].y = 90 - 45 * rsin(lat) / (0.4F + rcos(lat));
	}
	ProjectRectangular(p);
	return p->n;
}

local int PROJAPI ProjectFoucaut(const MAPPROJECT* p)
{
	// Foucaut (p.113)
	const int rx = ((p->rx + p->scroll) % 24 - 12) * 15;
	const int ry = p->ry * 30 - 90;
	int i;
	for (i = 0; i < p->n; i++)
	{
		const float lat = deg2rad(ry + p->src[i].y);
		p->dst[i].x = 180 + rcos(lat) * rsqr(rcos(lat / 2)) * (rx + p->src[i].x);
		p->dst[i].y = 90 + 90 * rtan(lat / 2);
	}
	return p->n;
}

local int PROJAPI ProjectCollignon1(const MAPPROJECT* p)
{
	// Collignon I (p.114)
	const int rx = ((p->rx + p->scroll) % 24 - 12) * 15;
	const int ry = 90 - p->ry * 30;
	int i;
	for (i = 0; i < p->n; i++)
	{
		const float lat = deg2rad(ry - p->src[i].y);
		const float a = rsqrts((1 - rsin(lat)) / 2);
		p->dst[i].x = 180 + a * (rx + p->src[i].x);
		p->dst[i].y = 180 * a;
	}
	return p->n;
}

local int PROJAPI ProjectCollignon2(const MAPPROJECT* p)
{
	// Collignon II (p.114)
	const int rx = ((p->rx + p->scroll) % 24 - 12) * 15;
	const int ry = 90 - p->ry * 30;
	const int dx = (ry > 0) ? 1 : -1;
	int i;
	for (i = 0; i < p->n; i++)
	{
		const float lat = deg2rad(ry - p->src[i].y) * dx;
		const float a = rsqrts(1 - rsin(lat));
		p->dst[i].x = 180 + a * (rx + p->src[i].x);
		p->dst[i].y = 90 + 90 * (a - 1) * dx;
	}
	return p->n;
}

local int PROJAPI ProjectNell(const MAPPROJECT* p)
{
	// Nell (p.115)
	const int rx = ((p->rx + p->scroll) % 24 - 12) * 15;
	const int ry = 90 - p->ry * 30;
	int i;
	for (i = 0; i < p->n; i++)
	{
		const float theta = SolveNell(deg2rad(ry - p->src[i].y));
		p->dst[i].x = 180 + (1 + rcos(theta)) * (rx + p->src[i].x) / 2;
		p->dst[i].y = 90 - 81.37F * theta;
	}
	return p->n;
}

local int PROJAPI ProjectPolyconic(const MAPPROJECT* p)
{
	// Polyconic (p.119)
	const int rectangular = p->projection->parameter;
	const int r = rectangular ? 47 : 37;
	const int rx = ((p->rx + p->scroll) % 24 - 12) * 15;
	const int ry = 90 - p->ry * 30;
	float x, y;
	int i;
	for (i = 0; i < p->n; i++)
	{
		const float lon = deg2rad(rx + p->src[i].x);
		const float lat = deg2rad(ry - p->src[i].y);
		const float rlat = rabs(lat);
		if (rlat < 0.000001F)
		{
			// Equator
			x = lon;
			y = 0;
		}
		else if (rlat > 1.5707963F)
		{
			// Poles
			x = 0;
			y = lat;
		}
		else
		{
			// Elsewhere
			float theta = lon * rsin(lat);
			if (rectangular)
				theta = 2 * ratan(theta, 2);
			x = rsin(theta) / rtan(lat);
			y = lat + (1 - rcos(theta)) / rtan(lat);
		}
		p->dst[i].x = 180 + r * x;
		p->dst[i].y = 90 - r * y;
	}
	return p->n;
}

local int PROJAPI ProjectAiry(const MAPPROJECT* p)
{
	// Oblique Airy at 30 degrees (p.127)
	static const float b = -0.057626074408164445782163683098253F;
	static const BYTE skip[] = { 12, 12, 12, 9, 8, 7 };
	int i = (p->rx + p->scroll) % 24 - 12;
	const int rx = i * 15;
	const int ry = 90 - p->ry * 30;
	float x, y, r;
	ASSERT(p->ry < dimof(skip));
	if ((i < -skip[p->ry]) || (i >= skip[p->ry]))
		return 0;
	for (i = 0; i < p->n; i++)
	{
		const float lon = deg2rad(rx + p->src[i].x);
		const float lat = deg2rad(ry - p->src[i].y);
		const float cosz = sin_30 * rsin(lat) + cos_30 * rcos(lat) * rcos(lon);
		ASSERT(cosz <= 1);
		if (cosz < 1)
		{
			const float t = (1 + cosz) / 2;
			ASSERT(t > 0);
			r = -rln(t) / (1 - cosz) - b / t;
		}
		else
			r = 0.5F - b;
		ASSERT(r > 0);
		x = r * rcos(lat) * rsin(lon);
		y = r * (cos_30 * rsin(lat) - sin_30 * rcos(lat) * rcos(lon));
		r = rsqr(x) + rsqr(y);
		if (r > 1)
		{
			// Fit to circle
			r = rsqrt(r);
			ASSERT(r > 0);
			x /= r;
			y /= r;
		}
		p->dst[i].x = 180 + 90 * x;
		p->dst[i].y = 90 - 90 * y;
	}
	return p->n;
}

local int PROJAPI ProjectLittrow(const MAPPROJECT* p)
{
	// Littrow (p.135)
	const int rx = ((p->rx + p->scroll) % 24 - 12) * 15;
	const int ry = 90 - p->ry * 30;
	int i;
	if ((rx < -90) || (rx >= 90))
		return 0;
	for (i = 0; i < p->n; i++)
	{
		const float lon = deg2rad(rx + p->src[i].x);
		const float lat = deg2rad(rclamp(ry - p->src[i].y, -80, 80));
		p->dst[i].x = 180 + 37 * rsin(lon) / rcos(lat);
		p->dst[i].y = 90 - 37 * rcos(lon) * rtan(lat);
	}
	return p->n;
}

local int PROJAPI ProjectStar(const MAPPROJECT* p)
{
	// Star (p.139)
	static const BYTE star[2][24] =
	{
		{ 1,1,7,7,7,13,13,13,19,19,19,25,25,25,31,31,31,37,37,37,43,43,43,49 },
		{ 1,1,1,11,11,11,11,11,20,20,20,20,29,29,29,29,29,39,39,39,39,39,49,49 }
	};
	static const float twist[2] = { -7.5F, 30.0F };
	int i = (p->rx + 11) % 24;
	const int rx = i * 15;
	const int ry = p->ry * 30;
	const float apex = deg2rad(star[p->projection->parameter][i] * 7.5F);
	const float rotate = deg2rad(p->scroll * 15 + twist[p->projection->parameter]);
	ASSERT((p->projection->parameter >= 0) && (p->projection->parameter < dimof(star)));
	for (i = 0; i < p->n; i++)
	{
		float theta = deg2rad(rx + p->src[i].x);
		float r = (ry + p->src[i].y) / 2;
		if (r >= 45)
		{
			// Converge to a point
			const float h = rsin(theta - apex) * (90 - r);
			theta = rasin(h / r) + apex;
		}
		theta += rotate;
		p->dst[i].x = 180 + r * rsin(theta);
		p->dst[i].y = 90 + r * rcos(theta);
	}
	return p->n;
}

local int PROJAPI ProjectCycloidal(const MAPPROJECT* p)
{
	// Jervis (p.144)
	const int rx = ((p->rx + p->scroll) % 24 - 12) * 15;
	const int ry = p->ry * 30;
	int i;
	if (ry > 90)
		return 0;
	for (i = 0; i < p->n; i++)
	{
		const float lon = deg2rad(rx + p->src[i].x);
		const float r = deg2rad(ry + p->src[i].y) * 55;
		p->dst[i].x = 180 + r * (lon + rsin(lon));
		p->dst[i].y = r * (1 + rcos(lon));
	}
	return p->n;
}

local int PROJAPI ProjectInterrupted(const MAPPROJECT* p)
{
	// Interrupted sinusoidal and Mollweide (p.168)
	static const BYTE interruption[5][2][24] =
	{
		{
			// No interruptions
			{ 120,120,120,120,120,120,120,120,120,120,120,120,120,120,120,120,120,120,120,120,120,120,120,120 },
			{ 120,120,120,120,120,120,120,120,120,120,120,120,120,120,120,120,120,120,120,120,120,120,120,120 }
		},
		{
			// Simple
			{ 53,53,53,53,53,53,53,53,53,160,160,160,160,160,160,160,160,160,160,160,160,160,160,160 },
			{ 12,12,12,12,12,80,80,80,80,80,80,133,133,133,133,133,133,220,220,220,220,220,220,220 }
		},
		{
			// Homolosine land
			{ 53,53,53,53,53,53,53,53,53,140,140,140,140,140,140,140,140,140,140,140,140,140,140,140 },
			{ 12,12,12,12,12,80,80,80,80,80,80,133,133,133,133,133,133,214,214,214,214,214,214,214 }
		},
		{
			// Homolosine ocean
			{ 30,30,30,30,30,30,116,116,116,116,116,116,116,116,116,116,210,210,210,210,210,210,210,210 },
			{ 50,50,50,50,50,50,50,50,138,138,138,138,138,138,138,138,138,138,138,217,217,217,217,217 }
		},
		{
			// Eumorphic
			{ 53,53,53,53,53,53,53,53,53,53,160,160,160,160,160,160,160,160,160,160,160,160,160,160 },
			{ 26,26,26,26,26,80,80,80,80,80,80,133,133,133,133,133,133,214,214,214,214,214,214,214 }
		},
	};
	static const BYTE offset[5] = { 0, 0, 0, 11, 0 };
	const int layout = p->projection->parameter / 10;
	const int equation = p->projection->parameter % 10;
	int i = (p->rx + p->scroll + offset[layout]) % 24;
	const int rx = i * 15;
	const int ry = 90 - p->ry * 30;
	const float apex = interruption[layout][p->ry / 3][i] * 1.5F;
	ASSERT((layout >= 0) && (layout < dimof(interruption)));
	for (i = 0; i < p->n; i++)
	{
		float lon = rx + p->src[i].x - apex;
		float lat = ry - p->src[i].y;
		float theta = deg2rad(lat);
		float a = apex;
		switch (equation)
		{
			case 0:
				// Sinusoidal
				break;
			case 1:
				// Mollweide
				theta = SolveMollweide(theta);
				lat = rsin(theta) * 90;
				break;
			case 2:
				// Homolosine
				if (rabs(lat) >= dms2deg(40, 44, 11.98F))
				{
					// Polar regions
					theta = SolveMollweide(theta);
					lat = rsin(theta) * 90 - 7.870838889F * rsgn(lat);
					lon *= 0.9F;
				}
				lat *= 1.095834887F;
				break;
			case 3:
				// Eumorphic
				{
					const float theta2 = SolveMollweide(theta);
					if ((layout != 0) && (rx >= 150) && (lat >= 40))
					{
						// Upper Eurasian landmass has a different meridian
						a -= 24.2F;
						lon += 30;
					}
					lon *= 2.11072073454F * rcos(theta2) / (rcos(theta2) + 1.11072073454F * rcos(theta));
					lat = 30 * (theta + rsin(theta2) / sin_45);
				}
				break;
			default:
				UNREACHABLE();
		}
		p->dst[i].x = lon * rcos(theta) + a;
		p->dst[i].y = 90 - lat;
	}
	return p->n;
}

local int PROJAPI ProjectMiller(const MAPPROJECT* p)
{
	// Miller I-III (p.179)
	const int ry = 90 - p->ry * 30;
	int i;
	for (i = 0; i < p->n; i++)
	{
		const float lat = deg2rad(ry - p->src[i].y);
		switch (p->projection->parameter)
		{
			case 1:
				// Perspective compromise
				p->dst[i].y = 90 - 52.72078F * (rsin(lat / 2) + rtan(lat / 2));
				break;
			case 2:
				// Modified Mercator (a)
				p->dst[i].y = 90 - 68.339314575F * rln(rtan(pi / 4 + lat / 3));
				break;
			case 3:
				// Modified Mercator (b)
				p->dst[i].y = 90 - 48.840578F * rln(rtan(pi / 4 + lat * 0.4F));
				break;
			default:
				UNREACHABLE();
		}
	}
	ProjectRectangular(p);
	return p->n;
}

local int PROJAPI ProjectArdenClose(const MAPPROJECT* p)
{
	// Arden-Close (p.186)
	const int rx = ((p->rx + p->scroll + 7) % 24) * 15;
	const int ry = 90 - p->ry * 30;
	const int dx = (rx < 180) ? 0 : 4;
	int i;
	if ((rx >= 90) && (rx < 270))
		return 0;
	for (i = 0; i < p->n; i++)
	{
		const float lon = deg2rad(rx + p->src[i].x);
		const float lat = deg2rad(ry - p->src[i].y);
		const float theta = (rabs(lat) < 0.000001F) ? 0 : (rabs(lat) < 1.5707963F) ? ratan(rtan(lat), rcos(lon)) : lat;
		const float x1 = lon / pi * 2 - dx;
		const float y1 = rsin(lat);
		const float x2 = rcos(lat) * rsin(lon);
		const float y2 = theta / pi * 2;
		ASSERT((x1 > -1.001F) && (x1 < 1.001F));
		ASSERT((y1 > -1.001F) && (y1 < 1.001F));
		ASSERT((x2 > -1.001F) && (x2 < 1.001F));
		ASSERT((y2 > -1.001F) && (y2 < 1.001F));
		p->dst[i].x = 180 + (x1 + x2) * 45;
		p->dst[i].y = 90 - (y1 + y2) * 45;
	}
	return p->n;
}

local int PROJAPI ProjectNellHammer(const MAPPROJECT* p)
{
	// Nell-Hammer (p.190)
	const int rx = ((p->rx + p->scroll) % 24 - 12) * 15;
	const int ry = 90 - p->ry * 30;
	int i;
	for (i = 0; i < p->n; i++)
	{
		const float lon = rx + p->src[i].x;
		const float lat = deg2rad(ry - p->src[i].y);
		p->dst[i].x = 180 + lon * (1 + rcos(lat)) / 2;
		p->dst[i].y = 90 - 157.67445545F * (lat - rtan(lat / 2));
	}
	return p->n;
}

local int PROJAPI ProjectEckert(const MAPPROJECT* p)
{
	// Eckert (pp.191)
	const int rx = ((p->rx + p->scroll) % 24 - 12) * 15;
	const int ry = 90 - p->ry * 30;
	float x = 0;
	int i;
	for (i = 0; i < p->n; i++)
	{
		float lon = rx + p->src[i].x;
		float lat = ry - p->src[i].y;
		switch (p->projection->parameter)
		{
			case 1:
				// Eckert I
				x = 1 - rabs(lat) / 180;
				break;
			case 2:
				// Eckert II
				x = rsqrt(4 - 3 * rsin(rabs(deg2rad(lat)))) / 2;
				lat = (1 - x) * rsgn(lat) * 180;
				break;
			case 3:
				// Eckert III
				x = (1 + rsqrts(1 - rsqr(lat / 90))) / 2;
				break;
			case 4:
				// Eckert IV
				x = SolveEckert4(deg2rad(lat));
				lat = rsin(x) * 90;
				x = (1 + rcos(x)) / 2;
				break;
			case 5:
				// Eckert V
				x = (1 + rcos(deg2rad(lat))) / 2;
				break;
			case 6:
				// Eckert VI
				x = SolveEckert6(deg2rad(lat));
				lat = x / pi * 180;
				x = (1 + rcos(x)) / 2;
				break;
			default:
				UNREACHABLE();
		}
		p->dst[i].x = 180 + lon * x;
		p->dst[i].y = 90 - lat;
	}
	return p->n;
}

local int PROJAPI ProjectWinkel(const MAPPROJECT* p)
{
	// Winkel I & II (pp.195)
	const int rx = ((p->rx + p->scroll) % 24 - 12) * 15;
	const int ry = 90 - p->ry * 30;
	float x;
	int i;
	for (i = 0; i < p->n; i++)
	{
		const float lon = rx + p->src[i].x;
		const float lat = ry - p->src[i].y;
		if (p->projection->parameter == 0)
			x = rcos(deg2rad(lat));
		else
			x = rsqrts(1 - rsqr(lat / 90));
		p->dst[i].x = 180 + lon * (x + 2 / pi) * 0.61101547F;
		p->dst[i].y = 90 - lat;
	}
	return p->n;
}

local int PROJAPI ProjectCraster(const MAPPROJECT* p)
{
	// Craster (p.200)
	const int rx = ((p->rx + p->scroll) % 24 - 12) * 15;
	const int ry = 90 - p->ry * 30;
	int i;
	for (i = 0; i < p->n; i++)
	{
		const float lon = rx + p->src[i].x;
		const float lat = deg2rad(ry - p->src[i].y) / 3;
		p->dst[i].x = 180 + lon * (2 * rcos(lat * 2) - 1);
		p->dst[i].y = 90 - 180 * rsin(lat);
	}
	return p->n;
}

local int PROJAPI ProjectKavrayskiy5(const MAPPROJECT* p)
{
	// Kavrayskiy V (p.200)
	static const float b = 0.73834093609048939243341540741457F;
	const int rx = ((p->rx + p->scroll) % 24 - 12) * 15;
	const int ry = 90 - p->ry * 30;
	int i;
	for (i = 0; i < p->n; i++)
	{
		const float lon = rx + p->src[i].x;
		const float lat = deg2rad(ry - p->src[i].y);
		p->dst[i].x = 180 + lon * rcos(lat) / rcos(lat * b);
		p->dst[i].y = 90 - 98.1764785378F * rsin(lat * b);
	}
	return p->n;
}

local int PROJAPI ProjectWagner(const MAPPROJECT* p)
{
	// Wagner (pp.203)
	const int rx = ((p->rx + p->scroll) % 24 - 12) * 15;
	const int ry = 90 - p->ry * 30;
	float theta = 0;
	int i;
	for (i = 0; i < p->n; i++)
	{
		const float lon = rx + p->src[i].x;
		float lat = ry - p->src[i].y;
		switch (p->projection->parameter)
		{
			case 1:
				// Wagner I
				theta = rasin(sin_60 * rsin(deg2rad(lat)));
				lat = theta / pi * 270;
				break;
			case 2:
				// Wagner II
				theta = rasin(0.88022F * rsin(0.8855F * deg2rad(lat)));
				lat = theta / pi * 270;
				break;
			case 3:
				// Wagner III
				theta = deg2rad(lat) / 1.5F;
				break;
			case 4:
				// Wagner IV
				theta = SolveWagner(2.9604205F * rsin(deg2rad(lat)));
				lat = 103.9229F * rsin(theta);
				break;
			case 5:
				// Wagner V
				theta = SolveWagner(3.00896F * rsin(0.8855F * deg2rad(lat)));
				lat = 103.9229F * rsin(theta);
				break;
			default:
				UNREACHABLE();
		}
		p->dst[i].x = 180 + lon * rcos(theta);
		p->dst[i].y = 90 - lat;
	}
	return p->n;
}

local int PROJAPI ProjectWagner6(const MAPPROJECT* p)
{
	// Wagner VI (p.205)
	const int rx = ((p->rx + p->scroll) % 24 - 12) * 15;
	const int ry = p->ry * 30;
	int i;
	for (i = 0; i < p->n; i++)
	{
		const float lon = rx + p->src[i].x;
		const float lat = ry + p->src[i].y;
		p->dst[i].x = 180 + lon * rsqrt(1 - rsqr(90 - lat) / 10800);
		p->dst[i].y = lat;
	}
	return p->n;
}

local int PROJAPI ProjectPutnins(const MAPPROJECT* p)
{
	// Putnins (pp.206)
	const int rx = ((p->rx + p->scroll) % 24 - 12) * 15;
	const int ry = 90 - p->ry * 30;
	float theta;
	int i;
	for (i = 0; i < p->n; i++)
	{
		float lon = rx + p->src[i].x;
		float lat = ry - p->src[i].y;
		switch (p->projection->parameter)
		{
			case 0:
				// Putnins P1
				lon *= rsqrt(4 - rsqr(lat) / 2700) - 1;
				break;
			case 1:
				// Putnins P2
				theta = SolvePutnins2(deg2rad(lat));
				lon *= rcos(theta) * 2 - 1;
				lat = rsin(theta) * 103.95112F;
				break;
			case 2:
				// Putnins P3
				lon *= 1 - rsqr(lat) / 8100;
				break;
			case 3:
				// Putnins P3'
				lon *= 1 - rsqr(lat) / 16200;
				break;
			case 4:
				// Putnins P4'
				theta = rasin(0.8838834765F * rsin(deg2rad(lat)));
				lon *= rcos(theta) / rcos(theta / 3);
				lat = rsin(theta / 3) * 254.55844F;
				break;
			case 5:
				// Putnins P5
				lon *= 2 - rsqrt(1 + rsqr(lat) / 2700);
				break;
			case 6:
				// Putnins P5'
				lon *= 1.5F - rsqrt(1 + rsqr(lat) / 2700) / 2;
				break;
			case 7:
				// Putnins P6
				theta = SolvePutnins6(4, 2.1471437182129378784F, deg2rad(lat));
				lon *= 2 - rsqrt(1 + rsqr(theta));
				lat = theta * 51.961524227F;
				break;
			case 8:
				// Putnins P6'
				theta = SolvePutnins6(6, 5.61125F, deg2rad(lat));
				lon *= (3 - rsqrt(1 + rsqr(theta))) / 2;
				lat = theta * 51.961524227F;
				break;
			default:
				UNREACHABLE();
		}
		p->dst[i].x = 180 + lon;
		p->dst[i].y = 90 - lat;
	}
	return p->n;
}

local int PROJAPI ProjectLoximuthal(const MAPPROJECT* p)
{
	// Siemon at 51.5 degrees [London] (p.208)
	const int rx = ((p->rx + p->scroll) % 24 - 12) * 15;
	const int ry = 90 - p->ry * 30;
	int i;
	for (i = 0; i < p->n; i++)
	{
		float lon = rx + p->src[i].x;
		float lat = ry - p->src[i].y;
		if (rabs(lat) > 89.999F)
			lon = 0;
		else
		{
			const float a = deg2rad(lat) - 0.898844564777F;
			if (rabs(a) < 0.000001F)
				lon *= 0.62251463663762F;
			else
				lon *= a / (rln(rtan(pi / 4 + deg2rad(lat) / 2)) - 1.05206568677F);
		}
		p->dst[i].x = 180 + lon;
		p->dst[i].y = 90 - lat;
	}
	return p->n;
}

local int PROJAPI ProjectQuarticAuthalic(const MAPPROJECT* p)
{
	// Siemon (p.209)
	const int rx = ((p->rx + p->scroll) % 24 - 12) * 15;
	const int ry = 90 - p->ry * 30;
	int i;
	for (i = 0; i < p->n; i++)
	{
		const float lon = rx + p->src[i].x;
		const float lat = deg2rad(ry - p->src[i].y);
		p->dst[i].x = 180 + lon * rcos(lat) / rcos(lat / 2);
		p->dst[i].y = 90 - 127.27922F * rsin(lat / 2);
	}
	return p->n;
}

local int PROJAPI ProjectMcBrydeThomas(const MAPPROJECT* p)
{
	// McBryde-Thomas I (p.210)
	static const float q = 1.36509F;
	const int rx = ((p->rx + p->scroll) % 24 - 12) * 15;
	const int ry = 90 - p->ry * 30;
	int i;
	for (i = 0; i < p->n; i++)
	{
		const float lon = rx + p->src[i].x;
		const float lat = deg2rad(ry - p->src[i].y);
		p->dst[i].x = 180 + lon * rcos(lat) / rcos(lat / q);
		p->dst[i].y = 90 - 98.5711722427F * rsin(lat / q);
	}
	return p->n;
}

local int PROJAPI ProjectTheTimes(const MAPPROJECT* p)
{
	// Moir (p.213)
	const int rx = ((p->rx + p->scroll) % 24 - 12) * 15;
	const int ry = 90 - p->ry * 30;
	int i;
	for (i = 0; i < p->n; i++)
	{
		const float t = rtan(deg2rad(ry - p->src[i].y) / 2);
		p->dst[i].x = 180 + (rx + p->src[i].x) * (1 - 0.46438F * rsqr(rsin(t * pi / 4)));
		p->dst[i].y = 90 - 90 * t;
	}
	return p->n;
}

local int PROJAPI ProjectRobinson(const MAPPROJECT* p)
{
	// Robinson (p.214)
	static const float cx[][4] =
	{
		{ 1.0000F,	-5.67239e-12F,	-7.15511e-05F,	 3.11028e-06F },
		{ 0.9986F,	-0.000482241F,	-2.48970e-05F,	-1.33094e-06F },
		{ 0.9954F,	-0.000831031F,	-4.48610e-05F,	-9.86588e-07F },
		{ 0.9900F,	-0.00135363F,	-5.96598e-05F,	 3.67749e-06F },
		{ 0.9822F,	-0.00167442F,	-4.49750e-06F,	-5.72394e-06F },
		{ 0.9730F,	-0.00214869F,	-9.03565e-05F,	 1.88767e-08F },
		{ 0.9600F,	-0.00305084F,	-9.00732e-05F,	 1.64869e-06F },
		{ 0.9427F,	-0.00382792F,	-6.53428e-05F,	-2.61493e-06F },
		{ 0.9216F,	-0.00467747F,	-0.000104566F,	 4.81220e-06F },
		{ 0.8962F,	-0.00536222F,	-3.23834e-05F,	-5.43445e-06F },
		{ 0.8679F,	-0.00609364F,	-0.000113900F,	 3.32521e-06F },
		{ 0.8350F,	-0.00698325F,	-6.40219e-05F,	 9.34582e-07F },
		{ 0.7986F,	-0.00755337F,	-5.00038e-05F,	 9.35532e-07F },
		{ 0.7597F,	-0.00798325F,	-3.59716e-05F,	-2.27604e-06F },
		{ 0.7186F,	-0.00851366F,	-7.01120e-05F,	-8.63072e-06F },
		{ 0.6732F,	-0.00986209F,	-0.000199572F,	 1.91978e-05F },
		{ 0.6213F,	-0.01041800F,	 8.83948e-05F,	 6.24031e-06F },
		{ 0.5722F,	-0.00906601F,	 0.000181999F,	 6.24033e-06F },
		{ 0.5322F,	 0.00000000F,	 0.000000000F,	 0.000000000F }
	};
	static const float cy[][4] =
	{
		{ 0.0000F,	 0.01240000F,	 3.72529e-10F,	 1.15484e-09F },
		{ 0.0620F,	 0.01240010F,	 1.76951e-08F,	-5.92321e-09F },
		{ 0.1240F,	 0.01239980F,	-7.09668e-08F,	 2.25753e-08F },
		{ 0.1860F,	 0.01240080F,	 2.66917e-07F,	-8.44523e-08F },
		{ 0.2480F,	 0.01239710F,	-9.99682e-07F,	 3.15569e-07F },
		{ 0.3100F,	 0.01241080F,	 3.73349e-06F,	-1.17790e-06F },
		{ 0.3720F,	 0.01235980F,	-1.39350e-05F,	 4.39588e-06F },
		{ 0.4340F,	 0.01255010F,	 5.20034e-05F,	-1.00051e-05F },
		{ 0.4968F,	 0.01231980F,	-9.80735e-05F,	 9.22397e-06F },
		{ 0.5571F,	 0.01203080F,	 4.02857e-05F,	-5.29010e-06F },
		{ 0.6176F,	 0.01203690F,	-3.90662e-05F,	 7.36117e-07F },
		{ 0.6769F,	 0.01170150F,	-2.80246e-05F,	-8.54283e-07F },
		{ 0.7346F,	 0.01135720F,	-4.08389e-05F,	-5.18524e-07F },
		{ 0.7903F,	 0.01090990F,	-4.86169e-05F,	-1.07180e-06F },
		{ 0.8435F,	 0.01034330F,	-6.46934e-05F,	 5.36384e-09F },
		{ 0.8936F,	 0.00969679F,	-6.46129e-05F,	-8.54894e-06F },
		{ 0.9394F,	 0.00840949F,	-0.000192847F,	-4.21023e-06F },
		{ 0.9761F,	 0.00616525F,	-0.000256001F,	-4.21021e-06F },
		{ 1.0000F,	 0.00000000F,	 0.000000000F,	 0.000000000F }
	};
	const int rx = ((p->rx + p->scroll) % 24 - 12) * 15;
	const int ry = 90 - p->ry * 30;
	const int dx = (ry > 0) ? 90 : -90;
	int i;
	for (i = 0; i < p->n; i++)
	{
		const float lon = rx + p->src[i].x;
		const float lat = ry - p->src[i].y;
		float a = rabs(lat);
		const int j = (int)(a / 5);
		ASSERT((j >= 0) && (j < dimof(cx)));
		a -= j * 5;
		p->dst[i].x = 180 + lon * (cx[j][0] + a * (cx[j][1] + a * (cx[j][2] + a * cx[j][3])));
		p->dst[i].y = 90 - dx * (cy[j][0] + a * (cy[j][1] + a * (cy[j][2] + a * cy[j][3])));
	}
	return p->n;
}

local int PROJAPI ProjectErdiKrausz(const MAPPROJECT* p)
{
	// Erdi Krauz (p.218)
	const int rx = ((p->rx + p->scroll) % 24 - 12) * 15;
	const int ry = 90 - p->ry * 30;
	int i;
	if ((ry == 90) || (ry == -60))
	{
		// Polar regions (Mollweide)
		const float cy = (ry == 90) ? 108.41052722F : 71.58947278F;
		for (i = 0; i < p->n; i++)
		{
            const float theta = SolveMollweide(deg2rad(ry - p->src[i].y));
			p->dst[i].x = 180 + 1.1143354F * (rx + p->src[i].x) * rcos(theta);
			p->dst[i].y = cy - 108.41052722F * rsin(theta);
		}
	}
	else
	{
		// Sinusoidal region
		for (i = 0; i < p->n; i++)
		{
            const float theta = rasin(0.8F * rsin(deg2rad(ry - p->src[i].y)));
			p->dst[i].x = 180 + (rx + p->src[i].x) * rcos(theta);
			p->dst[i].y = 90 - 83.93173F * theta;
		}
	}
	return p->n;
}

local int PROJAPI ProjectHatano(const MAPPROJECT* p)
{
	// Hatano (p.220)
	const int rx = ((p->rx + p->scroll) % 24 - 12) * 15;
	const int ry = 90 - p->ry * 30;
	const float a = (ry > 0) ? 122.32752F : 134.28697F;
	const float b = (ry > 0) ? 2.67595F : 2.43763F;
	int i;
	for (i = 0; i < p->n; i++)
	{
        const float theta = SolveHatano(b, deg2rad(ry - p->src[i].y));
		p->dst[i].x = 180 + (rx + p->src[i].x) * rcos(theta);
		p->dst[i].y = 91 - a * rsin(theta);
	}
	return p->n;
}

local int PROJAPI ProjectMecca(const MAPPROJECT* p)
{
	// Craig Mecca (p.228)
	static const float r = 97.69266F;
	const int rx = ((p->rx + p->scroll) % 24 - 15) * 15;
	const int ry = 90 - p->ry * 30;
	int i;
	if ((rx < -75) || (rx >= 75) || (ry < 0))
		return 0;
	for (i = 0; i < p->n; i++)
	{
		const float lon = deg2rad(rx + p->src[i].x);
		const float lat = deg2rad(ry - p->src[i].y);
		const float b = (rabs(lon) < 0.000001F) ? 1 : (lon / rsin(lon));
		p->dst[i].x = 180 + r * lon;
		p->dst[i].y = r - r * b * (rsin(lat) * rcos(lon) - rcos(lat) * 0.3955F);
	}
	return p->n;
}

local int PROJAPI ProjectRugby(const MAPPROJECT* p)
{
	// Rugby retroazimuthal [Actually London 51.5 degrees] (p.230)
	static const float sin_phi = 0.782608F;
	static const float cos_phi = 0.622515F;
	const int rx = ((p->rx + p->scroll) % 24 - 12) * 15;
	const int ry = 90 - p->ry * 30;
	int i;
	if ((rx < -90) || (rx >= 90))
		return 0;
	for (i = 0; i < p->n; i++)
	{
		const float lon = deg2rad(rx + p->src[i].x);
		const float lat = deg2rad(ry - p->src[i].y);
		float r = sin_phi * rsin(lat) + cos_phi * rcos(lat) * rcos(lon);
		r = rclamp(r, -0.999999F, 0.999999F);
		r = racos(r);
		ASSERT(rsin(r) > 0.000001F);
		r *= 57 / rsin(r);
		p->dst[i].x = 180 + r * cos_phi * rsin(lon);
		p->dst[i].y = 39 + r * (sin_phi * rcos(lat) - cos_phi * rsin(lat) * rcos(lon));
	}
	return p->n;
}

local int PROJAPI ProjectTripel(const MAPPROJECT* p)
{
	// Tripel (p.232)
	const int rx = ((p->rx + p->scroll) % 24 - 12) * 15;
	const int ry = 90 - p->ry * 30;
	int i;
	for (i = 0; i < p->n; i++)
	{
		const float lon = deg2rad(rx + p->src[i].x);
		const float lat = deg2rad(ry - p->src[i].y);
		const float a = racos(rcos(lat) * rcos(lon / 2));
		const float b = (rabs(a) < 0.000001F) ? 1 : (a / rsin(a));
		p->dst[i].x = 180 + 70 * (lon / pi + b * rcos(lat) * rsin(lon / 2));
		p->dst[i].y = 90 - 30 * (lat + b * rsin(lat));
	}
	return p->n;
}

local int PROJAPI ProjectWagner7(const MAPPROJECT* p)
{
	// Wagner VII (p.237)
	const int rx = ((p->rx + p->scroll) % 24 - 12) * 15;
	const int ry = 90 - p->ry * 30;
	int i;
	for (i = 0; i < p->n; i++)
	{
		const float lon3 = deg2rad(rx + p->src[i].x) / 3;
		const float lat = deg2rad(ry - p->src[i].y);
		const float theta = rasin(0.906307787F * rsin(lat));
		const float a = rcos(racos(rcos(theta) * rcos(lon3)) / 2);
		ASSERT(a > 0);
		p->dst[i].x = 180 + 180 * rcos(theta) * rsin(lon3) / a;
		p->dst[i].y = 90 - 77.28216F * rsin(theta) / a;
	}
	return p->n;
}

local int PROJAPI ProjectBriesemeister(const MAPPROJECT* p)
{
	// Briesemeister (p.239)
	const int rx = ((p->rx + p->scroll) % 24 - 12) * 15;
	const int ry = 90 - p->ry * 30;
	const int dx = (rx < 0) ? -1 : 1;
	int i;
	for (i = 0; i < p->n; i++)
	{
		const float lon = deg2rad(rx + p->src[i].x);
		const float lat = deg2rad(ry - p->src[i].y);
		const float sin_phi = rclamp((rsin(lat) - rcos(lat) * rcos(lon)) * sin_45, -1, 1);
		const float phi = rasin(sin_phi);
		const float cos_phi = rcos(phi);
		const float cos_lambda = rclamp((rsin(lat) + rcos(lat) * rcos(lon)) * sin_45 / cos_phi, -1, 1);
		const float lambda2 = racos(cos_lambda) / 2;
		const float r = 90 / rsqrt(1 + cos_phi * rcos(lambda2));
		ASSERT(r > 0);
		p->dst[i].x = 180 + 2 * r * cos_phi * rsin(lambda2) * dx;
		p->dst[i].y = 90 - r * sin_phi;
	}
	return p->n;
}

// The projection strings
LPCTSTR projection_name[Projection_Count][2];

void ProjectionDecompress(void)
{
#define PROJECTION(n, a, y, f, p, g)	_T(n) _T("\0") a
#define AUTHOR(a)						_T(a) _T("\0")
#define AUTHOR_NULL						_T("\0")
#define AUTHOR_PREV						_T("\001")
	LPCTSTR p =
#include "MapProjs.h"
	;
	LPCTSTR prev = NULL;
	int i;
	for (i = 0; i < Projection_Count; i++)
	{
		projection_name[i][0] = p;
		while (*(p++) != '\0');
		switch (*p)
		{
			case 0:
				// No author
				prev = NULL;
				p++;
				break;

			case 1:
				// Previous author
				p++;
				break;

			default:
				// Explicit name
				prev = p;
				while (*(p++) != '\0');
				break;
		}
		projection_name[i][1] = prev;
	}
	return;
#undef PROJECTION
}

// The projection table
#define PROJECTION(n, a, y, f, p, g)	{ (y * 2), (p), (g), (Project##f) },
#define CIRCA							1 |
const MAPPROJECTION projection[] =
{
#include "MapProjs.h"
};
